#!/usr/bin/env python

from msvcrt import getwch  # type: ignore

__all__ = ["get_char"]

NULL = "\0"
WIN_ESC = "\xe0"
KEYS_MAPPING = {
    "H": "up",
    "P": "down",
    "M": "right",
    "K": "left",
    "R": "insert",
    "S": "del",
    "I": "pgup",
    "Q": "pgdn",
}


def get_char() -> str:
    ch: str = getwch()
    if ch == WIN_ESC:  # When reading arrow/insert/del key, the first call returnx 0xe0
        ch = getwch()  # The second call returns the key code
        return KEYS_MAPPING.get(ch, NULL)
    return ch



#!/usr/bin/env python

import shutil
import sys
from typing import List, Optional, TextIO
from .ansi import (  # noqa
    ESC,
    NL,
    SPACE,
    CURSOR_SAVE_POS,
    CURSOR_RESTORE_POS,
    ERASE_LINE,
    RED,
    GREEN,
    YELLOW,
    BLUE,
    PURPLE,
    CYAN,
    WHITE,
    NORMAL,
    BOLD,
    NEGATIVE,
)

__all__ = ["Screen"]

DEFAULT_HEIGHT = 24
"Default screen height"
DEFAULT_WIDTH = 80
"Default screen width"


class Screen:
    def __init__(self, stream: TextIO = sys.stderr, fullscreen: bool = True, height: Optional[int] = None):
        """
        Initialize screen

        Args:
            stream: Output stream
            fullscreen: Full screen mode
            height: Screen height
        """
        self.stream: TextIO = stream
        self.data: List[str] = []
        self.fullscreen = fullscreen
        if self.fullscreen or height is None:
            self.height: int = self.get_terminal_height()
        else:
            self.height = min(height, self.get_terminal_height())
        # Save cursor position
        self.write(f"{CURSOR_SAVE_POS}")
        self.flush()

    @classmethod
    def get_terminal_height(cls) -> int:
        return shutil.get_terminal_size(fallback=(DEFAULT_WIDTH, DEFAULT_HEIGHT)).lines

    def write(self, line: str) -> None:
        "Add data to be written on the stream"
        self.data.append(line)

    def flush(self) -> None:
        "Write data to the stream and flush it"
        self.stream.write("".join(self.data))
        self.data = []
        self.stream.flush()

    def cleanup(self) -> None:
        "Clean screen and restore cursor position"
        self.erase_screen()
        if self.fullscreen:
            self.write(f"{CURSOR_RESTORE_POS}")
            self.move_up(self.height - 1)
        self.flush()

    def erase_screen(self) -> None:
        "Erase the screen"
        lines: int = self.height - 1
        self.erase_lines(lines)

    def erase_lines(self, lines: int) -> None:
        """
        Erase n lines

        Args:
            lines: number of lines to be erased
        """
        self.move_up(lines)
        self.write(f"{ERASE_LINE}{NL}" * lines)
        self.move_up(lines)

    def move_up(self, lines: int) -> None:
        """
        Move cursor up
        If the cursor is already at the edge of the screen, this has no effect.

        Args:
            lines: number of lines
        """
        self.write(f"{ESC}[{lines}A")

    def move_down(self, lines: int) -> None:
        """
        Move cursor down
        If the cursor is already at the edge of the screen, this has no effect.

        Args:
            lines: number of lines
        """
        self.write(f"{ESC}[{lines}B")



from pzp.finder import Finder
from pzp.screen import Screen
import pyte


class FakeIO:
    def __init__(self, stream):
        self.stream = stream

    def write(self, s):
        if s:
            return self.stream.feed(s.replace("\n", "\r\n"))
        else:
            return 0

    def flush(self):
        pass


class FakeTerminal:
    def __init__(self):
        self.screen = pyte.Screen(80, 24)
        self.stream = pyte.Stream(self.screen)
        self.output_stream = FakeIO(self.stream)

    @property
    def cursor(self):
        return self.screen.cursor

    def __str__(self):
        return "\n".join(self.screen.display)


def test_screen():
    terminal = FakeTerminal()
    screen = Screen(stream=terminal.output_stream, fullscreen=True, height=24)
    assert terminal.cursor.y == 0
    screen.move_down(10)
    screen.flush()
    assert terminal.cursor.y == 10
    screen.move_up(5)
    screen.flush()
    assert terminal.cursor.y == 5


def test_term():
    terminal = FakeTerminal()
    output_stream = terminal.output_stream
    assert terminal.cursor.y == 0

    lines = 4
    for i in range(0, lines):
        output_stream.write(f"line{i}>\n")
    assert terminal.cursor.y == lines

    candidates = [f"0x{x}" for x in range(0, 100)]
    finder = Finder(candidates=candidates, height=10, fullscreen=True, output_stream=output_stream)
    finder.setup()
    assert finder.selected == 0

    finder.process_key("down")
    finder.apply_filter()
    finder.update_screen()
    assert finder.selected == 1

    finder.process_key("9")
    finder.apply_filter()
    finder.update_screen()
    assert finder.selected == 1

    finder.process_key("9")
    finder.apply_filter()
    finder.update_screen()
    assert finder.selected == 0

    assert finder.prepare_result() == "0x99"
    # print(terminal)
    # assert terminal.cursor.y == finder.height + lines - 1

    finder.screen.cleanup()
    # output_stream.write("done\n")
    # print(terminal.cursor.y)
    # print("-----")
    # print(terminal)
    # print("-----")



#!/usr/bin/env python

try:
    from .input_posix import get_char
except ImportError:
    try:
        from .input_win import get_char
    except ImportError:
        raise Exception("Unsupported platform")

__all__ = ["get_char"]



#!/usr/bin/env python

import sys
from enum import Enum
from collections import ChainMap
from typing import Any, Callable, Iterator, Dict, Optional, Sequence, TextIO, Union
from .input import get_char
from .keys import ACTIONS, KEYS
from .ansi import (  # noqa
    NL,
    SPACE,
    ERASE_LINE,
    BLACK,
    RED,
    GREEN,
    YELLOW,
    BLUE,
    PURPLE,
    CYAN,
    WHITE,
    BLACK_BG,
    RED_BG,
    GREEN_BG,
    YELLOW_BG,
    BLUE_BG,
    PURPLE_BG,
    CYAN_BG,
    WHITE_BG,
    NORMAL,
    BOLD,
    NEGATIVE,
)
from .screen import Screen

__all__ = [
    "Finder",
    "Layout",
    "InfoStyle",
    "DEFAULT_POINTER",
    "DEFAULT_PROMPT",
]

DEFAULT_POINTER = ">"
"Default pointer"
DEFAULT_PROMPT = ">"
"Default input prompt"


class Confirm(Exception):
    pass


class Cancel(Exception):
    pass


class Layout(Enum):
    "Finder layouts"

    REVERSE_LIST = "reverse-list"
    " Display from the top of the screen, prompt at the bottom "


class InfoStyle(Enum):
    "Display style of finder info"

    DEFAULT = "default"
    " Display on the next line to the prompt "
    HIDDEN = "hidden"
    " Do not display finder info"


class Finder:
    def __init__(
        self,
        candidates: Union[Callable[[], Sequence[Any]], Iterator[Any], Sequence[Any]],
        fullscreen: bool = True,
        height: Optional[int] = None,
        format_fn: Callable[[Any], str] = lambda x: str(x),
        layout: Layout = Layout.REVERSE_LIST,
        info_style: InfoStyle = InfoStyle.DEFAULT,
        pointer_str: str = DEFAULT_POINTER,
        prompt_str: str = DEFAULT_PROMPT,
        output_stream: TextIO = sys.stderr,
    ):
        """
        Initializate Finder object

        Args:
            candidates: Candidates
            fullscreen: Full screen mode
            height: Finder window height
            format_fn: Items format function
            layout: Finder layout
            info_style: Determines the display style of finder info
            pointer_str: Pointer to the current line
            prompt_str: Input prompt
        """
        self.fullscreen = fullscreen
        self.height = height
        self.format_fn = format_fn
        self.layout: Layout = layout
        self.info_style: InfoStyle = info_style
        self.pointer_str = pointer_str
        self.no_pointer_str = " " * len(pointer_str)
        self.prompt_str = prompt_str
        self.output_stream = output_stream
        self.keycodes_actions: Dict[str, str] = dict(ChainMap(*[{KEYS[v]: k for v in vlist} for k, vlist in ACTIONS.items()]))
        # Get the candidates
        if isinstance(candidates, Iterator) or callable(candidates):
            self.get_items_fn: Union[None, Callable[[], Sequence[Any]], Iterator[Any]] = candidates
            self.candidates: Sequence[Any] = []
        else:
            self.get_items_fn = None
            self.candidates = candidates

    def show(self, input: Optional[str] = None) -> Any:
        """
        Open pzp and return the selected element

        Args:
            input: initial search string

        Returns:
            item: the selected item
        """
        self.setup(input=input)
        try:
            while True:
                self.process_key(get_char())
                self.apply_filter()
                self.update_screen()
        except Confirm:
            return self.prepare_result()
        except Cancel:
            return None
        finally:
            self.screen.cleanup()

    @property
    def screen_items(self) -> Sequence[Any]:
        "Candidates to be displayed on the screen"
        return self.matching_candidates[self.offset : self.offset + self.max_candidates]

    @property
    def screen_items_len(self) -> int:
        "Number of items on the screen"
        return len(self.screen_items)

    @property
    def candidates_len(self) -> int:
        "Number of candidates"
        return len(self.candidates)

    @property
    def matching_candidates_len(self) -> int:
        "Number of matching candidates"
        return len(self.matching_candidates)

    @property
    def info_lines(self) -> int:
        "Number of info lines"
        return 1 if self.info_style == InfoStyle.DEFAULT else 0

    @property
    def prompt_lines(self) -> int:
        "Number of prompt lines"
        return len(self.prompt_str.split(f"{NL}"))

    @property
    def margin_lines(self) -> int:
        "Screen margin"
        return self.info_lines + self.prompt_lines

    @property
    def max_candidates(self) -> int:
        "Maximun number of candidates printables on the screen"
        return self.screen.height - self.margin_lines

    def setup(self, input: Optional[str] = None) -> None:
        """
        Setup Finder execution

        Args:
            input: initial search string
        """
        self.input: str = input or ""
        # Load the candidate list
        self.refresh_candidates()
        # Calculate the required height and setup the screen
        height = self.height if self.height is not None else self.candidates_len + self.margin_lines
        self.screen: Screen = Screen(stream=self.output_stream, fullscreen=self.fullscreen, height=height)
        # Filter the items, calculate the screen offset
        self.apply_filter()
        self.update_screen(erase=False)

    def refresh_candidates(self) -> None:
        "Load/reload the candidate list"
        # Get items
        if isinstance(self.get_items_fn, Iterator):
            self.candidates = list(self.get_items_fn)
        elif callable(self.get_items_fn):
            self.candidates = list(self.get_items_fn())
        # Reset selected/offset
        self.selected: int = 0
        self.offset: int = 0

    def process_key(self, ch: str) -> None:
        "Process the pressed key"
        action = self.keycodes_actions.get(ch)
        if action == "accept":  # Confirm
            raise Confirm
        elif action == "abort":  # Cancel
            raise Cancel
        elif action == "down":  # Move one line down
            self.selected = self.selected + 1
        elif action == "up":  # Move one line up
            self.selected = self.selected - 1
        elif action == "page-down":  # Move one page down
            self.selected = self.selected + self.max_candidates
        elif action == "page-up":  # Move one page up
            self.selected = self.selected - self.max_candidates
        elif action == "backward-delete-char":  # Delete one characted
            if self.input:
                self.input = self.input[:-1]
        elif action == "ignore":  # Skip
            pass
        elif ch >= SPACE:  # Append the character to line
            self.input = self.input + ch

    def apply_filter(self) -> None:
        "Filter the items, calculate the screen offset"
        self.matching_candidates: Sequence[Any] = list(filter(self.match, self.candidates))
        # Adject selected
        self.selected = max(min(self.selected, self.matching_candidates_len - 1), 0)
        # Calculate the offset
        if self.selected >= self.offset + self.max_candidates:
            self.offset = self.selected - self.max_candidates + 1
        elif self.selected < self.offset:
            self.offset = self.selected
        if self.offset < 0:
            self.offset = 0

    def update_screen(self, erase: bool = True) -> None:
        "Update the screen - erase the old items, print the filtered items and the prompt"
        if erase:
            self.screen.erase_screen()
        if self.layout == Layout.REVERSE_LIST:
            self.print_items()
            self.print_empty_lines()
            self.print_info()
            self.print_prompt()
        self.screen.flush()

    def print_items(self) -> None:
        for i, item in enumerate(self.screen_items):
            is_selected = i + self.offset == self.selected
            if is_selected:
                self.screen.write(
                    f"{ERASE_LINE}{RED}{BOLD}{BLACK_BG}{self.pointer_str}{NORMAL} {BOLD}{self.format_fn(item)}{NORMAL}{NL}"
                )
            else:
                self.screen.write(f"{ERASE_LINE}{BLACK_BG}{self.no_pointer_str}{NORMAL} {self.format_fn(item)}{NL}")

    def print_empty_lines(self) -> None:
        if self.fullscreen:
            lines = self.max_candidates - self.screen_items_len
        else:
            lines = min(self.candidates_len, self.max_candidates - self.screen_items_len)
        self.screen.write(f"{NL}" * lines)

    def print_info(self) -> None:
        "Print info"
        if self.info_style == InfoStyle.DEFAULT:
            self.screen.write(f"  {ERASE_LINE}{YELLOW}{self.matching_candidates_len}/{self.candidates_len}{NORMAL}{NL}")

    def print_prompt(self) -> None:
        "Print prompt"
        self.screen.write(f"{ERASE_LINE}{CYAN}{self.prompt_str}{NORMAL} {self.input}")

    def match(self, item: Any) -> bool:
        return self.input.lower() in self.format_fn(item).lower()

    def prepare_result(self) -> Any:
        "Output the selected item, if any"
        try:
            return self.matching_candidates[self.selected]
        except IndexError:
            return None



#!/usr/bin/env python

__all__ = ["get_char"]

import sys
import termios
import tty

NULL = "\0"
ESC = "\u001b"


def get_char() -> str:
    try:
        fd = sys.stdin.fileno()
        attrs = termios.tcgetattr(fd)
        tty.setraw(fd)
        ch = sys.stdin.read(1)
        if ch == ESC:
            ch = sys.stdin.read(1)
            if ch == "[":
                ch = sys.stdin.read(1)
                if ch == "A":
                    ch = "up"
                elif ch == "B":
                    ch = "down"
                elif ch == "C":
                    ch = "right"
                elif ch == "D":
                    ch = "left"
                elif ch == "2":
                    ch = "insert"
                    sys.stdin.read(1)  # skip ~
                elif ch == "3":
                    ch = "del"
                    sys.stdin.read(1)  # skip ~
                elif ch == "5":
                    ch = "pgup"
                    sys.stdin.read(1)  # skip ~
                elif ch == "6":
                    ch = "pgdn"
                    sys.stdin.read(1)  # skip ~
                else:
                    ch = NULL
        return ch
    finally:
        termios.tcsetattr(fd, termios.TCSAFLUSH, attrs)



def test_import_module():
    import pzp

    assert pzp


def test_import_ansi():
    from pzp import ansi

    assert ansi


def test_import_finder():
    from pzp import finder

    assert finder


def test_import_input():
    from pzp import input

    assert input


def test_import_keys():
    from pzp import keys

    assert keys


def test_import_screen():
    from pzp import screen

    assert screen
